
from colito.logging import getModuleLogger
from sergio.kernels.gramian import GramianWithNullspace, GramianFromArray

log = getModuleLogger('')
log.setLevel(10)

import sys
import os

from collections import namedtuple
ExperimentNames = namedtuple('ExperimentNames', ('result','log','tag'))
def get_file_names(base_dir, rank, comparison, depth, prefix, approximation_factor, gamma,prediciser,top, iteration):
    spred = f'p{prediciser}' if prediciser else ''
    stop = f'k{top}' if top!=1 else ''
    siter = f'i{iteration}' if iteration else ''
    tag = f'{prefix}{comparison}-r{rank}d{depth}a{approximation_factor*100:03.0f}g{gamma*100:04.0f}{stop}{spred}{siter}'
    
    fname_log = os.path.join(base_dir, f'output/logs/{tag}.log')
    fname_out = os.path.join(base_dir, f'output/results/{tag}.json')
    return ExperimentNames(log=fname_log,result=fname_out,tag=tag)

def get_result_subgroup_descriptions(file):
    import json
    with open(file, 'rb') as fid:
        js = json.load(fid)
    sgs = [sg['description'] for sg in js['subgroups']]
    return sgs


def main(K, D, rank, comparison, depth, prefix='', approximation_factor=1., gamma=1.,prediciser='',top=1, removed_subgroups=[], base_dir='.', iteration=0):
    from sergio.computation import Computation
    from sergio import FileManager
    
    names = get_file_names(base_dir=base_dir, rank=rank, comparison=comparison, depth=depth, prefix=prefix, approximation_factor=approximation_factor, gamma=gamma,prediciser=prediciser,top=top, iteration=iteration)

    if os.path.exists(names.result):
        raise FileExistsError(f'Refuse to overwrite output file: {names.result}')
    import numpy as np

    log.rlim.delay['PROGRESS'] = 30
    log.add_file(names.log)
    
    c = Computation(file_manager=FileManager(paths={'data':'data'}))

    c.dataset = D
    c.prediciser = prediciser
    c.language = 'conjunctions'
    
    g_rank = max(10,rank)
    z = lambda x: x-x.sum()/len(x)
    if removed_subgroups:
        c.removed_subgroups = removed_subgroups
        NS = np.c_[[z(sg.validity) for sg in c.removed_subgroups]].T.astype(float)
        G = GramianWithNullspace(K, nullspace=NS, rank=g_rank)
    else:
        G = GramianFromArray(K, rank=g_rank)
    
    c.gramian = G
    
    import sergio.scores.kernels
    c.load_measure('mmd', rank=rank, comparison=comparison, gamma=gamma)
    if c.removed_subgroups:
        c.measure.evaluate(c.removed_subgroups[0])
    c.load_optimistic_estimator('mmd-sd', max_rank=rank, comparison=comparison, gamma=gamma)
    
    c.optimise_inplace(depths=depth, approximation_factor=approximation_factor,k=top)

    js = c.summarise_to_json()
    with open(names.result,'w') as fid:
        fid.write(js)
    return c

        
def load_dataset(name):
    if name == 'twitter':
        from loaders.twitter import load_twitter
        loader = load_twitter
    elif name == 'stocks':
        from loaders.stocks import load_stocks
        loader = load_stocks
    elif name == 'chem':
        from loaders.chem import load_chem
        loader = load_chem
    else:
        log.error(f'Unknown dataset: {name}. Available datasets: twitter, stocks, chem.')
        
    return loader()
        
    
if __name__ == '__main__':
    import sys
    import argparse
    ap = argparse.ArgumentParser('experiment runner')
    ap.add_argument('-D', '--dataset')
    ap.add_argument('-r', '--rank', default=1,type=int)
    ap.add_argument('-d', '--depth', default=2,type=int)
    ap.add_argument('-g', '--gamma', default=1.,type=float)
    ap.add_argument('-a', '--approximation-factor', dest='approximation_factor', default=1.,type=float)
    ap.add_argument('-c','--comparison',required=True)
    ap.add_argument('-p','--prediciser',default='')
    ap.add_argument('-t', '--tag', default='')
    ap.add_argument('-k', '--top', default=1,type=int)
    ap.add_argument('-R', '--remove-result', nargs='*',default=[])
    ap.add_argument('-i', '--iterations', type=int, default=1, help='Iterative discovery')
    ap.add_argument('-b', '--base_dir', default='.', help='Make all paths relative to this dir.')
    
    argv = sys.argv[1:]
    if len(sys.argv) == 2 and sys.argv[1] == '--use-static-arguments':
        argv = '-Dstocks -ccontrastive -d1 -ttest -p5:slabs_positive:ranged'.split(' ')
        argv = '-Dstocks -ccontrastive -d1 -ttest1 -p5:slabs_positive:ranged -Routput/results/test-stocks-contrastive-r1d1a100g0100p5:slabs_positive:ranged.json'.split(' ')
        argv = '-Dstocks -ccontrastive -d2 -ttestit -p5:slabs_positive:ranged -i3'.split(' ')
        argv = '-Dtwitter -canomaly -r1 -d3 -g1 -p5:slabs_positive:ranged -i3 -ttestit -r100'.split(' ')
    nmsp = ap.parse_args(argv)
    
    prefix = '-'.join(([nmsp.tag] if nmsp.tag else [])+[nmsp.dataset,''])
    if nmsp.remove_result:
        remove_subgroups = []
        import json
        for result in nmsp.remove_result:
            sgs = get_result_subgroup_descriptions(result)
            print(f'Adding {len(sgs)} subgroups to remove list: {sgs}', file=sys.stderr)
            remove_subgroups += sgs
        
    else:
        remove_subgroups = []
    
    from types import SimpleNamespace
    params = SimpleNamespace(
        rank=nmsp.rank,
        comparison=nmsp.comparison,
        depth=nmsp.depth,
        approximation_factor=nmsp.approximation_factor,
        gamma=nmsp.gamma,
        prefix= prefix,
        prediciser=nmsp.prediciser,
        top=nmsp.top,
        base_dir = nmsp.base_dir,
    )
    log.add_stderr()
    eas, K = load_dataset(nmsp.dataset)
    remove_subgroups = list(nmsp.remove_result)
    try:
        for iteration in range(nmsp.iterations):
            try:
                c = main(K=K, D=eas, **params.__dict__, iteration=iteration, removed_subgroups=remove_subgroups)
                sgs = [str(s) for s in c.subgroups]
                log.info(f'Iteration {iteration+1}/{nmsp.iterations}: Finished with results: {sgs}')
            except FileExistsError:
                sgs = None
            if sgs is None:
                names = get_file_names(**params.__dict__, iteration=iteration)
                sgs = get_result_subgroup_descriptions(names.result)
                log.info(f'Iteration {iteration+1}/{nmsp.iterations}: Previously computed with results: {sgs}')
            remove_subgroups += list(sgs)
            log.info(f'After iteration: {iteration+1}/{nmsp.iterations}: Current remove list {len(remove_subgroups)}: {remove_subgroups}.')
    except Exception as e:
        log.exception(f'Error {e}')
